/*
	*
	* Copyright 2019-2021 Britanicus <marcusbritanicus@gmail.com>
	*
	* This file is a part of DesQ project ( https://gitlab.com/DesQ/ )
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* at your option, any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, write to the Free Software
	* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
	* MA 02110-1301, USA.
	*
*/

#include "PowerManager.hpp"
#include "PowerManagerTray.hpp"
#include "UPowerBackend.hpp"

#include <desq/Settings.hpp>

DesQ::Settings *desqSett;

QIcon makeIcon( double pc, bool ac ) {
    /*
        *
        * pc -> Battery Percentage
        * ac -> AC Power Connected
        *
    */

    if ( ac and pc > 99.0 )
        return QIcon( ":/icons/desq-power-manager.png" );

    QColor border;

    double mFraction = pc / 100.0;
    int red = 0, green = 0;

    // Critical: Bright Red
    if ( pc < 5.0 ) {
        // Bright Red
        red = 0xff;
        green = 0x00;
    }

    // Change from Red to Yellow: When mFraction = 0.05, green = 0; mFraction = 0.4, green = 0xBB
    else if ( mFraction < 0.4 ) {
        red = 0xBB;
        green = ( int )( ( mFraction - 0.05 ) * 550 );
    }

    // Remain Yellow
    else if ( mFraction <= 0.6 ) {
        red = 0xBB;
        green = 0xBB;
    }

    // Change from Yellow to Green: When totalF = 0.6, red = 0xBB; mFraction = 1, red = 0;
    else {
        red = ( int )( ( 1 - mFraction ) * 637.5 );
        green = 0xBB;
    }

    border = QColor( red, green, 0 );

    /* Battery charge gradient */
    QLinearGradient gradient( QPoint( 100, 0 ), QPoint( 220, 0 ) );
    gradient.setColorAt( 0.0, border.darker( 120 ) );
    gradient.setColorAt( 0.4, border );
    gradient.setColorAt( 0.6, border );
    gradient.setColorAt( 1.0, border.darker( 120 ) );
    QBrush fill = QBrush( gradient );

    /* Clean the image */
    QImage img( 320, 320, QImage::Format_ARGB32 );
    QPainter p( &img );
    p.setRenderHints( QPainter::Antialiasing );
    img.fill( Qt::transparent );

    /* Border */
    p.save();
    p.setPen( QPen( border, 10, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin ) );
    p.drawRoundedRect( QRect( 90, 30, 140, 280 ), 10, 10 );
    p.setPen( Qt::NoPen );
    p.setBrush( border );
    p.drawRoundedRect( QRect( 120, 5, 80, 30 ), 10, 10 );
    p.restore();

    /* Charge of the battery */
    p.save();
    p.setPen( Qt::NoPen );
    p.setBrush( fill );
    p.drawRoundedRect( QRect( 100, 40 + 260 * ( 100 - pc ) / 100.0, 120, 260 * pc / 100.0 ), 5, 5 );
    p.restore();

    /* AC Power Connected: Charging */
    qDebug() << "AC Power Supply connected" << ac;
    if ( ac ) {
        QPainterPath pp;
        pp.moveTo( QPoint( 200, 60 ) );
        pp.lineTo( QPoint( 120, 170 ) );
        pp.lineTo( QPoint( 200, 170 ) );
        pp.lineTo( QPoint( 120, 280 ) );

        p.save();
        p.setPen( QPen( QColor( "#CCCC00" ), 20, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin ) );
        p.setBrush( QColor( "#AABBBB00" ) );
        p.drawPath( pp );
        p.restore();
    }

    else if ( pc < 5 and not ac ) {
        /** Draw a red exclamation point */
        p.save();
        p.setPen( QPen( QColor( Qt::darkRed ), 30, Qt::SolidLine, Qt::RoundCap, Qt::RoundJoin ) );
        p.drawLine( QPointF( 160, 60 ), QPointF( 160, 200 ) );
        p.drawLine( QPointF( 160, 240 ), QPointF( 160, 250 ) );
        p.restore();
    }

    p.end();

    return QIcon( QPixmap::fromImage( img ) );
};

DesQPowerManager::DesQPowerManager() {

    desqSett = new DesQ::Settings( "PowerManager" );
    qDebug() << "Settings";

    desqPower = new UPowerBackend();
    qDebug() << "Backend";

    connect( desqPower, &UPowerBackend::lidOpened,            this, &DesQPowerManager::handleLidOpened );
    connect( desqPower, &UPowerBackend::lidClosed,            this, &DesQPowerManager::handleLidClosed );

    connect( desqPower, &UPowerBackend::switchedToBattery,    this, &DesQPowerManager::handleSwitchedToBattery );
    connect( desqPower, &UPowerBackend::switchedToACPower,    this, &DesQPowerManager::handleSwitchedToACPower );

    connect( desqPower, &UPowerBackend::batteryChargeChanged, this, &DesQPowerManager::handleBatteryChargeChanged );
    connect( desqPower, &UPowerBackend::batteryFullyCharged,  this, &DesQPowerManager::handleBatteryFullyCharged );
    connect( desqPower, &UPowerBackend::batteryNearlyEmpty,   this, &DesQPowerManager::handleBatteryNearlyEmpty );
    connect( desqPower, &UPowerBackend::batteryEmpty,         this, &DesQPowerManager::handleBatteryEmpty );

    qDebug() << "Connections";

    /** Start giving out signals */
    desqPower->startManagement();
    qDebug() << "Start startManagement";

    onBattery = desqPower->onBattery();
    qDebug() << "OnBattery" << onBattery;

    // For testing purposes
    // double pc = 0;
    // QTimer *timer = new QTimer();
    // timer->start( 100 );
    // connect(
    //     timer, &QTimer::timeout, [=]() mutable {
    //         handleBatteryChargeChanged( pc );
    //         pc += 0.5;
    //
    //         if ( pc >= 100.0 )
    //             timer->stop();
    //     }
    // );
};

/* Start the tray icon */
void DesQPowerManager::startTray() {

    // tray = new PowerTrayIcon( this );
    tray = new QSystemTrayIcon( this );
    tray->show();

    handleBatteryChargeChanged( desqPower->batteryCharge() );
};

void DesQPowerManager::notifyUser( QString title, QString message, short weight ) {

    /** Always show critical notifications */
    if ( not desqSett->value( "ShowNotifications" ) and weight != 2 )
        return;

    QString urgency;
    switch( weight ) {
        case 0:
            urgency = "low";
            break;
        case 1:
            urgency = "normal";
            break;
        default:
            urgency = "critical";
            break;
    }

    QProcess::startDetached(
        "desq-notifier", {
            "-i", "desq", "-a", "DesQ Power Manager",
            "-t", "5000", "-u", urgency, "-r", mID,
            title, message
        }
    );
};

/* Lid Open/Close Signals */
void DesQPowerManager::handleLidOpened() {
    // Nothing required as of now: Wayfire handles this with swaylock + swayidle
};

void DesQPowerManager::handleLidClosed() {
    // Nothing required as of now: Wayfire handles this with swaylock + swayidle
};

/* Battery Charge/Discharge Signals */
void DesQPowerManager::handleSwitchedToBattery() {
    // 1. Send a notification using notify-send.
    // 2. Set/Restore display brightness
    // 3. Load/Unload power profiles

    QMap<QString, Device*> devices = desqPower->devices();
    for( QString dev: devices.keys() ) {
        devices[ dev ]->update();

        if ( devices[ dev ]->type == Device::DeviceBattery ) {
            if ( devices[ dev ]->state == Device::DeviceStateDischarging ) {
                qDebug() << "Device running on battery.";
                qDebug() << devices[ dev ]->name << "is discharging.";
                qDebug() << "Estimated time remaining" << devices[ dev ]->timeToEmpty << "seconds";
            }
        }

        else if ( devices[ dev ]->type == Device::DeviceUps ) {
            qDebug() << "Device running on UPS.";
            qDebug() << devices[ dev ]->name << "is discharging.";
            qDebug() << "Estimated time remaining" << devices[ dev ]->timeToEmpty << "seconds";
        }
    }

    qDebug() << "Switched to Battery";
    onBattery = true;

    notifyUser(
        "AC Disconnected",
        "The AC Power supply has been disconnected. You're running on battery backup."
    );

    handleBatteryChargeChanged( desqPower->batteryCharge() );
};

void DesQPowerManager::handleSwitchedToACPower() {
    // 1. Send a notification using notify-send.
    // 2. Set/Restore display brightness
    // 3. Load/Unload power profiles

    qDebug() << "Switched to AC Power";
    onBattery = false;

    notifyUser(
        "AC Connected",
        "The AC Power supply has been connected. Your battery is charging."
    );

    handleBatteryChargeChanged( desqPower->batteryCharge() );
};

/* Battery charge changed */
void DesQPowerManager::handleBatteryChargeChanged( double value ) {
    // When discharging
    //  1. Set the display brightness based on percentage power left.
    //  2. oad/Unload suitable power modules.
    // When charging, change the battery icon.

    qDebug() << "[Handle BCC] Battery charge:" << value << onBattery;
    tray->setIcon( makeIcon( value, not onBattery ) );
};

/* Battery full signal */
void DesQPowerManager::handleBatteryFullyCharged() {
    // Display a notification for the user

    qDebug() << "Charging complete";
    notifyUser( "Battery Charged", "Your battery/UPS is completely charged." );
};

/* Battery nearly (5m) empty */
void DesQPowerManager::handleBatteryNearlyEmpty() {
    // Inform the user that around 5 minutes of battery charge remain.
    // Need to show full-screen blocking

    tray->setIcon( makeIcon( 5, false ) );
    notifyUser(
        "Battery Nearly Empty",
        "Your battery is nearly empty with approximately 5 minutes of charge is remaining. "
        "Save your work immediately and find an alternate source of power supply.",
        2
    );
};

/* Battery almost (1m) empty */
void DesQPowerManager::handleBatteryEmpty() {
    // Inform the user that around 1 minute of battery charge remains.
    // Prepare for suspend/shutdown
    // Need to show full-screen blocking

    tray->setIcon( makeIcon( 0, false ) );
    notifyUser(
        "Battery Critical",
        "Your battery is has under 1 minute of charge is remaining. "
        "Your system will be suspended immediately unless a power supply is connected.",
        2
    );
};
